#ifndef __PLUGIN_TYPE_INFO__
#define __PLUGIN_TYPE_INFO__

namespace utility {

struct plugin_type_info;

template <typename _PLUGIN_TYPE>
const plugin_type_info& plugin_typeid(_PLUGIN_TYPE& _p)
{
	return _p.__get_plugin_typeid();
}

#define plugin_class_typeid(_PLUGIN_TYPE) (_PLUGIN_TYPE::__class_plugin_typeid())
#define plugin_class_pointer_typeid(_PLUGIN_POINTER_TYPE) (((_PLUGIN_POINTER_TYPE)0)->__class_plugin_typeid())


// Type information about plugin classes.
struct plugin_type_info
{
public:
	// For collation and comparison.  Does not correlate in any way to inheritance.
	bool operator==(const plugin_type_info& c) const	{ return _id == c._id; }
	bool operator!=(const plugin_type_info& c) const	{ return _id != c._id; }
	bool operator<(const plugin_type_info& c) const		{ return _id < c._id; }
	bool operator<=(const plugin_type_info& c) const	{ return _id <= c._id; }
	bool operator>(const plugin_type_info& c) const		{ return _id > c._id; }
	bool operator>=(const plugin_type_info& c) const	{ return _id >= c._id; }

	// Returns true if this type is the given type (i.e. the class with this type-info is the class or
	// derives from the class with the given type-info).
	bool is(const plugin_type_info& info) const	{ int temp; return below_or_equal(info, temp); }

	// Reserved for internal use.
public:
	struct _parent_node
	{
		inline _parent_node(const plugin_type_info& _info, int _ptroffset, const _parent_node* _next = 0) : _info(_info), _ptroffset(_ptroffset), _next(_next)  {}
		const plugin_type_info& _info;
		int _ptroffset;
		const _parent_node* _next;
	};
	const void* cast_pointer(const void* obj, const plugin_type_info& from, const plugin_type_info& to) const  { return cast_pointer((void*)obj, from, to); }
	void* cast_pointer(void* obj, const plugin_type_info& from, const plugin_type_info& to) const
	{
		int fromoffset = 0;
		if (!below_or_equal(from, fromoffset))  return 0;
		int tooffset = 0;
		if (!below_or_equal(to, tooffset))  return 0;
		return (void*)((char*)obj - fromoffset + tooffset);
	}
protected:
	plugin_type_info(unsigned int _id, const _parent_node* _pn) : _id(_id), _pn(_pn)  {}
	bool below_or_equal(const plugin_type_info& info, int& ptroffset) const
	{
		if (_id == info._id)  return true;
		else
		{
			for (const _parent_node* curnode = _pn; curnode != 0; curnode = curnode->_next)
			{
				if (curnode->_info.below_or_equal(info, ptroffset))
				{
					ptroffset += curnode->_ptroffset;
					return true;
				}
			}
			return false;
		}
	}
	unsigned int _id;
	const _parent_node* _pn;
};

// Dynamically casts an plugin class to another plugin class.  The syntax for this operation is identical to
// dynamic_cast, except it only works with pointers, not references.
template <typename _TO, typename _FROM>
_TO plugin_cast(_FROM _obj)
{
		if (_obj == 0) return (_TO)0;
		else
		{
			return static_cast<_TO>(plugin_typeid(*_obj).cast_pointer(_obj,
				plugin_class_pointer_typeid(_FROM), plugin_class_pointer_typeid(_TO)));
		}
}

#define __CALC_POINTER_OFFSET(_t1, _t2) ((int)(void*)(_t2*)(_t1*)(void*)16777216 - (int)16777216)
#define __DECLARE_PLUGIN_PARENT_NODE(_name, _pn, _p, _next)  static const utility::plugin_type_info::_parent_node _pn(_p::__class_plugin_typeid(), __CALC_POINTER_OFFSET(_name, _p), _next);
#define __PLUGIN_IMPLEMENTATION_ID(_name)  ((unsigned int)_name::__class_plugin_typeid)
#define __PLUGIN_INTERFACE_ID(_id)  ((unsigned int)_id * 4 + 3)

#define __PRE_DECLARE_PI(_name) \
virtual const utility::plugin_type_info& __get_plugin_typeid() const  { return __class_plugin_typeid(); } \
struct __my_plugin_type_info : public utility::plugin_type_info { inline __my_plugin_type_info(int _id, const _parent_node* _pn) : plugin_type_info(_id, _pn) {} }; \
inline static const utility::plugin_type_info& __class_plugin_typeid() \
{

#define __POST_DECLARE_PI(_id, _name, _pn) \
	static const __my_plugin_type_info ret(_id, _pn); \
	return ret; \
}

#define __DECLARE_PI0(_id, _name) \
typedef _name __root_plugin_type; \
__PRE_DECLARE_PI(_name) \
__POST_DECLARE_PI(_id, _name, 0)

#define __DECLARE_PI1(_id, _name, _p1) \
typedef _p1::__root_plugin_type __root_plugin_type; \
__PRE_DECLARE_PI(_name) \
	__DECLARE_PLUGIN_PARENT_NODE(_name, pn1, _p1, 0); \
__POST_DECLARE_PI(_id, _name, &pn1)

#define __DECLARE_PI2(_id, _name, _p1, _p2) \
typedef _p1::__root_plugin_type __root_plugin_type; \
__PRE_DECLARE_PI(_name) \
	__DECLARE_PLUGIN_PARENT_NODE(_name, pn2, _p2, 0); \
	__DECLARE_PLUGIN_PARENT_NODE(_name, pn1, _p1, &pn2); \
__POST_DECLARE_PI(_id, _name, &pn1)

#define __DECLARE_PI3(_id, _name, _p1, _p2, _p3) \
typedef _p1::__root_plugin_type __root_plugin_type; \
__PRE_DECLARE_PI(_name) \
	__DECLARE_PLUGIN_PARENT_NODE(_name, pn3, _p3, 0); \
	__DECLARE_PLUGIN_PARENT_NODE(_name, pn2, _p2, &pn3); \
	__DECLARE_PLUGIN_PARENT_NODE(_name, pn1, _p1, &pn2); \
__POST_DECLARE_PI(_id, _name, &pn1)

#define __DECLARE_PI4(_id, _name, _p1, _p2, _p3, _p4) \
typedef _p1::__root_plugin_type __root_plugin_type; \
__PRE_DECLARE_PI(_name) \
	__DECLARE_PLUGIN_PARENT_NODE(_name, pn4, _p4, 0); \
	__DECLARE_PLUGIN_PARENT_NODE(_name, pn3, _p3, &pn4); \
	__DECLARE_PLUGIN_PARENT_NODE(_name, pn2, _p2, &pn3); \
	__DECLARE_PLUGIN_PARENT_NODE(_name, pn1, _p1, &pn2); \
__POST_DECLARE_PI(_id, _name, &pn1)


// Macros for declaring classes with the native run-time information.  This declaration assumes that
// _id is an unsigned int unique amongst all of the native classes declared, and deterministic across
// compilers.  The former implies that every native declaration has to know a little bit about every
// other native declaration (in order to avoid id conflicts).  The latter implies that the id cannot
// be compiler or load-time dependent (so pointers won't work!).  And one additional restriction, _id
// must be less than INT_MAX / 4, or it may conflict with other native ids.  This macro must be defined
// under a public class heading (couldn't put public: in a macro).
// Syntax: __DECLARE_NATIVE<N>(<unique id>, <class name>, <1st parent>, <2nd parent>, ...), 
// where <N> is the number of parents the class has.
#define __PLUGIN_INTERFACE0(_id, _name)						__DECLARE_PI0(__PLUGIN_INTERFACE_ID(_id), _name) 
#define __PLUGIN_INTERFACE1(_id, _name, _p1)				__DECLARE_PI1(__PLUGIN_INTERFACE_ID(_id), _name, _p1) 
#define __PLUGIN_INTERFACE2(_id, _name, _p1, _p2)			__DECLARE_PI2(__PLUGIN_INTERFACE_ID(_id), _name, _p1, _p2) 
#define __PLUGIN_INTERFACE3(_id, _name, _p1, _p2, _p3)		__DECLARE_PI3(__PLUGIN_INTERFACE_ID(_id), _name, _p1, _p2, _p3) 
#define __PLUGIN_INTERFACE4(_id, _name, _p1, _p2, _p3, _p4)	__DECLARE_PI4(__PLUGIN_INTERFACE_ID(_id), _name, _p1, _p2, _p3, _p4) 

#define __PLUGIN_IMPLEMENTATION0(_name)						__DECLARE_PI0(__PLUGIN_IMPLEMENTATION_ID(_name), _name) 
#define __PLUGIN_IMPLEMENTATION1(_name, _p1)				__DECLARE_PI1(__PLUGIN_IMPLEMENTATION_ID(_name), _name, _p1) 
#define __PLUGIN_IMPLEMENTATION2(_name, _p1, _p2)			__DECLARE_PI2(__PLUGIN_IMPLEMENTATION_ID(_name), _name, _p1, _p2) 
#define __PLUGIN_IMPLEMENTATION3(_name, _p1, _p2, _p3)		__DECLARE_PI3(__PLUGIN_IMPLEMENTATION_ID(_name), _name, _p1, _p2, _p3) 
#define __PLUGIN_IMPLEMENTATION4(_name, _p1, _p2, _p3, _p4)	__DECLARE_PI4(__PLUGIN_IMPLEMENTATION_ID(_name), _name, _p1, _p2, _p3, _p4) 


}; //namespace utility {



#endif  //#ifndef __PLUGIN_TYPE_INFO__
